from rlist import rlist

from micromanager.config.app import app_config
from micromanager.models import Project
from micromanager.commands.errors import ArgumentValidationError


def parse_projects(projects: list[str]) -> list[Project]:
    """
    Parses a list of strings into Project models.
    Raises ArgumentValidationError if a project is not in the current system.
    """
    current_system = app_config.get_current_system()
    current_project_names = rlist(current_system.projects).map(lambda p: p.name)

    invalid_input = rlist(projects).select(lambda p: p not in current_project_names)
    if len(invalid_input) > 0:
        msg = f"Cannot start projects {invalid_input} as they are not part of the current system '{current_system.name}'.\nAvailable projects: {current_project_names}"
        raise ArgumentValidationError(msg)

    _projects = [p for p in current_system.projects if p.name in projects]
    return _projects


def get_project_names() -> rlist[str]:
    """
    Get the names of the projects in the current system.
    """
    return app_config.get_current_system().projects.map(lambda p: p.name)



from typing import Generator
from collections import Counter

from rich import print

from micromanager.config.app import app_config
from micromanager.commands.app import app
from micromanager.compose.ps import DockerComposePs, DockerComposePsResponse


@app.command()
def status() -> None:
    """
    Print the status of all configured projects in the current system.
    """
    response = DockerComposePs.call(app_config.get_current_system().projects)

    table = StatusTable(response)
    print(table.format())


class StatusTable:
    HEADERS = ("PROJECT", "STATUS")
    ROW_LENGTHS = (25,)

    def __init__(self, response: DockerComposePsResponse) -> None:
        self._response = response

    def format(self) -> str:
        table = self._format_headers() + "\n"
        table_len = len(self._response)
        for i, row in enumerate(self._rows()):
            table += self._format_row(row)

            if i != table_len - 1:
                table += "\n"

        return table

    def _format_headers(self) -> str:
        headers = self.HEADERS[0]
        headers += self._whitespace(self.HEADERS[0])
        headers += self.HEADERS[1]

        return headers

    def _format_row(self, row: tuple[str, str]) -> str:
        _row = row[0]
        _row += self._whitespace(row[0])
        _row += row[1]
        return _row

    def _whitespace(self, value: str) -> str:
        return " " * self._whitespace_chars(value)

    def _whitespace_chars(self, value: str) -> int:
        return self.ROW_LENGTHS[0] - len(value)

    def _rows(self) -> Generator[tuple[str, str]]:
        for project in self._response:
            if len(project.containers) == 0:
                status = "STOPPED"
            else:
                stata = Counter(project.containers.map(lambda c: c.status))

                status = stata.most_common(1)[0][0]
                in_status = stata[status]
                fraction = "(" + str(in_status) + "/" + str(stata.total()) + ")"

                status = status.value + " " + fraction

            yield (project.project_name, status)



from typing import Annotated

from typer import Argument
from rich import print

from micromanager.config.app import app_config
from micromanager.commands.app import app
from micromanager.commands.errors import ArgumentValidationError


def _autocompletion(incomplete: str):
    names = app_config.systems.keys()
    for name in names:
        if name.startswith(incomplete):
            yield name


@app.command()
def use(system: Annotated[str, Argument(autocompletion=_autocompletion)]) -> None:
    """
    Set the given system as your current working system.
    """
    s = app_config.systems.get(system, None)
    if s is None:
        available_systems = list(app_config.systems.keys())
        raise ArgumentValidationError(
            f"system {system} is not found in the configured systems; check {app_config.config_file_path()}\nAvailable systems: {available_systems}"
        )

    app_config.set_current_system(s)
    print(f"Using system: {s.name}")



from dataclasses import dataclass
from pathlib import Path

from rlist import rlist


@dataclass(frozen=True, kw_only=True)
class Service:
    name: str


@dataclass(frozen=True, kw_only=True)
class Project:
    name: str
    compose_file_path: Path
    services: rlist[Service]


@dataclass(frozen=True, kw_only=True)
class System:
    name: str
    is_default: bool
    projects: rlist[Project]



import os
from pathlib import Path
from typing import Optional, Protocol
from dataclasses import replace

import json
import yaml
from rlist import rlist

from micromanager.models import System, Project, Service
from micromanager.config.models import ConfiguredSystem
from micromanager.config.errors import (
    ConfigFileDoesNotExistError,
    ComposeFileDoesNotExistError,
    InvalidConfigFileError,
)


class FileParser(Protocol):
    """
    A parser used to parse json or yaml files.
    """

    def load(self, path: Path) -> dict: ...


class JsonParser:
    """
    Parser for json files.
    Used to parse the micromanager configuration.
    """

    def __init__(self) -> None:
        self._json = json

    def load(self, path: Path) -> dict:
        """
        Parse and load the file in the given path in a dictionary.
        """
        with open(os.path.expandvars(path), "r") as f:
            try:
                parsed = self._json.load(f)
            except json.decoder.JSONDecodeError as e:  # ty: ignore unresolved-attribute
                raise InvalidConfigFileError(str(path), str(e)) from None

        return parsed


class YamlParser:
    """
    Parser for yaml files.
    Used to parse docker compose configurations.
    """

    def __init__(self) -> None:
        self._yaml = yaml

    def load(self, path: Path) -> dict:
        """
        Parse and load the file in the given path in a dictionary.
        """
        with open(os.path.expandvars(path), "r") as f:
            parsed = self._yaml.load(f, yaml.Loader)

        return parsed


class Parser:
    """
    Parser for micromanager configuration file.
    Responsible for parsing the json config of micromanager and the
    yml docker compose configs.
    Also, validates for correct application logic configuration.
    """

    def __init__(
        self,
        path: Path,
        json_parser: Optional[FileParser] = None,
        yaml_parser: Optional[FileParser] = None,
    ) -> None:
        self._path = path
        self._effective_path: Optional[Path] = None
        self._json = JsonParser() if json_parser is None else json_parser
        self._yaml = YamlParser() if yaml_parser is None else yaml_parser

    def parse(self) -> dict[str, System]:
        """Parse the configuration file into a dictionary."""
        if self._path.exists(follow_symlinks=False):
            self._effective_path = self._path
            return self._parse_config()

        raise ConfigFileDoesNotExistError(str(self._path))

    def _parse_config(self) -> dict[str, System]:
        json_file = self._json.load(
            self._effective_path  # ty: ignore invalid-argument-type
        )
        config = dict()

        if "systems" not in json_file.keys():
            raise InvalidConfigFileError(
                str(self._effective_path),
                "'systems' field does not exist in config.json",
            )

        systems = json_file["systems"]

        if not isinstance(systems, dict):
            raise InvalidConfigFileError(
                str(self._effective_path),
                "The value of the systems field is not a valid object",
            )

        for name, system in json_file["systems"].items():
            if not isinstance(system, dict):
                raise InvalidConfigFileError(
                    str(self._effective_path),
                    f"The system '{name}' is not a valid object",
                )
            config[name] = self._build_system(name, system)

        self._validate_config(config)
        config = {name: sys.to_system() for name, sys in config.items()}

        if len(config) == 1:
            config[name] = replace(config[name], is_default=True)

        return config

    def _build_system(self, name: str, attrs: dict) -> ConfiguredSystem:
        is_default = attrs.get("default", None)

        if "projects" not in attrs:
            raise InvalidConfigFileError(
                str(self._effective_path),
                f"'projects' field does not exist in the '{name}' system",
            )

        projects = rlist()
        for project_name, project_attrs in attrs["projects"].items():
            if not isinstance(project_attrs, dict):
                raise InvalidConfigFileError(
                    str(self._effective_path),
                    f"The project '{project_name}' of system '{name}' is not a valid object",
                )
            projects.append(self._build_project(project_name, project_attrs))

        system = ConfiguredSystem(name=name, is_default=is_default, projects=projects)
        return system

    def _build_project(self, name: str, attrs: dict) -> Project:
        if "compose_file_path" not in attrs:
            raise InvalidConfigFileError(
                str(self._effective_path),
                f"Project '{name}' does not contain a 'compose_file_path' field",
            )

        compose_file_path = Path(os.path.expandvars(attrs["compose_file_path"]))
        if not compose_file_path.exists():
            raise ComposeFileDoesNotExistError(name, str(compose_file_path))

        services = self._build_services(compose_file_path)

        project = Project(
            name=name, compose_file_path=compose_file_path, services=services
        )
        return project

    def _build_services(self, compose_file_path: Path) -> rlist[Service]:
        compose_file = self._yaml.load(compose_file_path)

        services = rlist([Service(name=s) for s in compose_file["services"]])
        return services

    def _validate_config(self, config: dict[str, System]) -> None:
        defaults = {sys_name for sys_name, sys in config.items() if sys.is_default}
        if len(defaults) > 1:
            raise InvalidConfigFileError(
                str(self._effective_path),
                f"More than one default systems configured ({defaults}); only one system can be the default",
            )

        if len(config) > 1 and len(defaults) == 0:
            raise InvalidConfigFileError(
                str(self._effective_path),
                'One system must be the default; set "default"="true"',
            )

        if len(config) == 1 and list(config.values())[0].is_default is False:
            sys_name = list(config.keys())[0]
            raise InvalidConfigFileError(
                str(self._effective_path),
                f'\'{sys_name}\' must be the default system since it is the only one. set "default"="true"',
            )



from dataclasses import dataclass
from typing import Optional

from rlist import rlist

from micromanager.models import System, Project


@dataclass
class ConfiguredSystem:
    """
    A system as defined in the configuration file with its primitive values,
    before it is casted into a domain model.
    """

    name: str
    is_default: Optional[bool]
    projects: rlist[Project]

    def to_system(self) -> System:
        """
        Convert the configuration model into a domain model with the provided
        values.
        """
        is_default = False if self.is_default is None else self.is_default
        return System(name=self.name, is_default=is_default, projects=self.projects)



from pathlib import Path
from typing import Optional
import os

from platformdirs import user_config_dir

from micromanager.models import System
from micromanager.config.parser import Parser
from micromanager.config.errors import NoDefaultSystemError


class AppConfig:
    """
    Configuration object for the micromanager tool.
    Defines helper methods to interact with the configuration.
    """

    def __init__(self, parser=None) -> None:
        self._path = None
        self._config: Optional[dict[str, System]] = None
        self._default_system: Optional[System] = None
        self._current_system: Optional[System] = None
        self._parser: Parser = (
            Parser(self.config_file_path()) if parser is None else parser
        )

    @property
    def systems(self) -> dict[str, System]:
        """
        Returns all configured systems as a dictionary with system name as key
        and system as value.
        """
        return self._get_config()

    def get_default_system(self) -> System:
        """
        Get the default micromanager system.
        If only one system is configured, that is the default one.
        """
        if self._default_system is None:
            self._default_system = self._get_default_system()

        return self._default_system

    def get_current_system(self) -> System:
        """
        Get the current system, i.e. the system that is currently selected.
        If a specific system is not selected, then the default system is set as the current system.
        All commands of the micromanager CLI are directed to the current system.
        """
        if self._current_system is None:
            self._current_system = self._get_current_system()

        return self._current_system

    def set_current_system(self, system: System) -> None:
        """
        Set the current system, i.e. the system that is currently selected.
        All commands of the micromanager CLI are directed to the current system.
        """
        self._current_system = system

    def config_file_path(self) -> Path:
        """
        Return the path of the micromanager configuration file.

        The path can be set from the enviroment variable MICROMANAGER_CONFIG_PATH.
        If not set it defaults to a sensible value according to the operating system.
        """
        if self._path is None:
            self._path = self._get_config_file_path()

        return self._path

    def _get_default_system(self) -> System:
        config = self._get_config()
        for system in config.values():
            if system.is_default:
                return system

        raise NoDefaultSystemError()

    def _get_current_system(self) -> System:
        if self._current_system is None:
            self._current_system = self.get_default_system()

        return self._current_system

    def _get_config(self) -> dict:
        if self._config is None:
            self._config = self._parser.parse()

        return self._config

    def _get_config_file_path(self) -> Path:
        env_var = os.getenv("MICROMANAGER_CONFIG_PATH", None)

        if env_var is not None:
            return Path(os.path.expandvars(env_var))

        return Path(user_config_dir("micromanager"), "config.json")



class ConfigError(Exception):
    """Base class for configuration errors."""


class NoDefaultSystemError(ConfigError):
    """
    No default system found in the configuration.

    Serves mainly as a placeholder; an error should have been raised in the parser level
    for this kind of situation before reaching the config level.
    """

    def __init__(self) -> None:
        msg = "No default system found in the configuration."
        super().__init__(msg)


class ParserError(ConfigError):
    """Base class for errors originating in the configuration parser."""

    pass


class InvalidConfigFileError(ParserError):
    """The configuration file is invalid, e.g. because of a typo."""

    def __init__(self, path: str, error: str) -> None:
        msg = f"The micromanager configuration file {path} is invalid: {error}"
        super().__init__(msg)


class ConfigFileDoesNotExistError(ParserError):
    """The micromanager configuration file was not found in the predefined paths."""

    def __init__(self, path: str) -> None:
        msg = f"The micromanager configuration file was not found in the predefined paths: {path}"
        super().__init__(msg)


class ComposeFileDoesNotExistError(ParserError):
    """A compose file was not found in the given path."""

    def __init__(self, project_name: str, path: str) -> None:
        msg = f"The compose_file_path field of the project {project_name} does not point to an existing path {path}"
        super().__init__(msg)



from enum import Enum
from dataclasses import dataclass
from typing import TypeAlias

from python_on_whales import DockerClient, Container as _Container
from python_on_whales.exceptions import DockerException
from rlist import rlist

from micromanager.models import Project
from micromanager.compose.errors import DockerComposePsError


class Status(Enum):
    CREATED = "CREATED"
    RUNNING = "RUNNING"
    RESTARTING = "RESTARTING"
    EXITED = "EXITED"
    PAUSED = "PAUSED"
    DEAD = "DEAD"


@dataclass
class Container:
    name: str
    status: Status


@dataclass
class DockerComposeProjectStatus:
    project_name: str
    containers: rlist[Container]


DockerComposePsResponse: TypeAlias = rlist[DockerComposeProjectStatus]


class DockerComposePs:
    """The docker compose ps command interface"""

    FLAGS = {
        "all": True,
    }

    @classmethod
    def call(cls, projects: rlist[Project]) -> DockerComposePsResponse:
        """
        Run the docker compose ps command.
        """
        response = rlist()
        for project in projects:
            client = DockerClient(compose_files=project.compose_file_path)
            try:
                containers = client.compose.ps(**cls.FLAGS)
            except DockerException as e:
                raise DockerComposePsError(str(e))

            response.append(
                DockerComposeProjectStatus(
                    project_name=project.name,
                    containers=cls._map_containers(containers),
                )
            )

        return response

    @classmethod
    def _map_containers(cls, containers: list[_Container]) -> rlist[Container]:
        mapped = [
            Container(c.name, cls._map_status(c.state.status)) for c in containers
        ]

        return rlist(mapped)

    @staticmethod
    def _map_status(status: str) -> Status:
        match status:
            case "created":
                return Status.CREATED
            case "running":
                return Status.RUNNING
            case "restarting":
                return Status.RESTARTING
            case "exited":
                return Status.EXITED
            case "paused":
                return Status.PAUSED
            case _:
                return Status.DEAD



from typing import Annotated

from typer import Argument
from rlist import rlist
from rich import print

from micromanager.compose.down import DockerComposeDown
from micromanager.config.app import app_config
from micromanager.commands.app import app
from micromanager.commands.utils import parse_projects, get_project_names


def _autocompletion(incomplete: str):
    for name in get_project_names():
        if name.startswith(incomplete):
            yield name


@app.command()
def stop(
    projects: Annotated[
        list[str] | None, Argument(autocompletion=_autocompletion)
    ] = None,
) -> None:
    """
    Stop the given projects by running compose down.
    If the projects argument is empty, stops all projects of the current system.
    """
    if projects is None:
        _projects = app_config.get_current_system().projects
    else:
        _projects = parse_projects(projects)

    _projects = rlist(_projects)
    DockerComposeDown.call(rlist(_projects))
    print(f"Stopped projects: {_projects.map(lambda p: p.name).to_list()}")



class DockerComposeError(Exception):
    """
    An error occured on the execution of a docker compose command.
    """


class DockerComposeUpError(DockerComposeError):
    """
    An error occured on the execution of docker compose up.
    """

    def __init__(self, projects: list[str], error: str) -> None:
        msg = (
            "An error occured during the execution of docker compose up for the"
            + f" projects {projects}: {error}"
        )
        super().__init__(msg)


class DockerComposeDownError(DockerComposeError):
    """
    An error occured on the execution of docker compose down.
    """

    def __init__(self, projects: list[str], error: str) -> None:
        msg = f"An error occured during the execution of docker compose down for the projects {projects}: {error}"
        super().__init__(msg)


class DockerComposePsError(DockerComposeError):
    """
    An error occured on the execution of docker compose ps.
    """


class DockerComposeRestartError(DockerComposeError):
    """
    An error occured on the execution of docker compose restart
    """



from .start import start as start
from .use import use as use
from .config import config as config
from .stop import stop as stop
from .status import status as status
from .restart import restart as restart
from .app import app

__all__ = ["app"]



from python_on_whales import DockerClient
from python_on_whales.exceptions import DockerException
from rlist import rlist

from micromanager.models import Project
from micromanager.compose.errors import DockerComposeUpError


class DockerComposeUp:
    """The docker compose up command interface"""

    FLAGS = {
        "detach": True,
        "quiet": False,
    }

    @classmethod
    def call(cls, projects: rlist[Project]) -> None:
        """
        Run the docker compose up command for the given projects.

        Runs compose up on each project separately according to the
        order of the given list.
        """
        for project in projects:
            client = DockerClient(compose_files=project.compose_file_path)
            try:
                client.compose.up(**cls.FLAGS)
            except DockerException as e:
                raise DockerComposeUpError(
                    projects.map(lambda p: p.name).to_list(), str(e)
                ) from None



from rich import print

from micromanager.models import System, Project
from micromanager.config.app import app_config
from micromanager.commands.app import app


@app.command()
def config() -> None:
    """
    Display the current configuration.
    """
    print(f"micromanager configuration file: {app_config.config_file_path()}")
    for name, system in app_config.systems.items():
        print(
            f"[b bright_green]{name}:[/b bright_green]\n\t{system_pretty_str(system)}"
        )


def system_pretty_str(system: System) -> str:
    """
    Return a pretty-formatted string representation of a System.
    """
    s = f"""name: {system.name}
        is_default: {system.is_default}
        projects:"""
    s += "\n"
    for project in system.projects:
        s += "\t\t" + project_pretty_str(project)

    return s


def project_pretty_str(project: Project) -> str:
    """
    Return a pretty-formatted string representation of a Project.
    """
    s = f"""name: {project.name}
                compose_file_path: {project.compose_file_path}
                services:"""
    s += "\n"
    for service in project.services:
        s += "\t\t\t" + service.name + "\n"

    return s
